#include "starwin.h"

#ifdef HAVE_MONO
#include "mono.h"
#endif

#ifdef HAVE_BPP4
#include "bpp4.h"
#endif

#ifdef HAVE_BPP8
#include "bpp8.h"
#endif

#ifdef HAVE_BPP15
#include "bpp15.h"
#endif

#ifdef HAVE_BPP16
#include "bpp16.h"
#endif

#ifdef HAVE_BPP24
#include "bpp24.h"
#endif

#ifdef HAVE_BPP32
#include "bpp32.h"
#endif

int swWidth = 640;
int swHeight = 480;
int swScanWidth = 640;
int bits_per_pixel = 1;
swCOLOR foreground={0,0,0,0};
swCOLOR background={255,255,255,0};

int (*drawpixel)(int x,int y);
int (*getpixel)(int x,int y);
int (*drawline)(int x1,int y1,int x2,int y2);
int (*setcolor)(int r,int g,int b);
int (*bitblt)(int x,int y,int w,int h,char *bmp);
int (*getbitmap)(int x,int y,int w,int h,char *bmp);
int (*miximage)(unsigned char *dest,unsigned char *src,int w,int h);
int (*scroll)(int x,int y,int w,int h,int line);
int (*fillbox)(int x,int y,int w,int h);
int (*xorbox)(int x,int y,int w,int h);
int (*drawfont)(int x,int y,int w,int h,unsigned char *font);
int (*setforeground)(int r,int g,int b);
int (*setbackground)(int r,int g,int b);
unsigned char * (*create_image_xpm)(int *w,int *h,char **xpm);
int (*mice_grinit)(char **arrow_xpm);
unsigned char * (*create_buffer)(int w,int h);

typedef struct {
  char * ModeStr;
  int    ModeId;
  int    bits;
} swModeInfo;

#define SUPPORT_MODES 44
swModeInfo swMI[SUPPORT_MODES]={

  {"640x480x2"      ,G640x480x2      , 1},
  {"720x348x2"      ,G720x348x2      , 1},

  {"320x200x16"     ,G320x200x16     , 4},
  {"320x200x256"    ,G320x200x256    , 8},
  {"320x200x32K"    ,G320x200x32K    ,15},
  {"320x200x64K"    ,G320x200x64K    ,16},
  {"320x200x16M"    ,G320x200x16M    ,24},
  {"320x200x16M32"  ,G320x200x16M32  ,32},
  
  {"640x480x16"     ,G640x480x16     , 4},
  {"640x480x256"    ,G640x480x256    , 8},
  {"640x480x32K"    ,G640x480x32K    ,15},
  {"640x480x64K"    ,G640x480x64K    ,16},
  {"640x480x16M"    ,G640x480x16M    ,24},
  {"640x480x16M32"  ,G640x480x16M32  ,32},
  
  {"800x600x16"     ,G800x600x16     , 4},
  {"800x600x256"    ,G800x600x256    , 8},
  {"800x600x32K"    ,G800x600x32K    ,15},
  {"800x600x64K"    ,G800x600x64K    ,16},
  {"800x600x16M"    ,G800x600x16M    ,24},
  {"800x600x16M32"  ,G800x600x16M32  ,32},
 
  {"1024x768x16"    ,G1024x768x16    , 4},
  {"1024x768x256"   ,G1024x768x256   , 8},
  {"1024x768x32K"   ,G1024x768x32K   ,15},
  {"1024x768x64K"   ,G1024x768x64K   ,16},
  {"1024x768x16M"   ,G1024x768x16M   ,24},
  {"1024x768x16M32" ,G1024x768x16M32 ,32},
  
  {"1152x864x16"   ,G1152x864x16     , 4},
  {"1152x864x256"  ,G1152x864x256    , 8},
  {"1152x864x32K"  ,G1152x864x32K    ,15},
  {"1152x864x64K"  ,G1152x864x64K    ,16},
  {"1152x864x16M"  ,G1152x864x16M    ,24},
  {"1152x864x16M32",G1152x864x16M32  ,32},

  {"1280x1024x16"   ,G1280x1024x16   , 4},
  {"1280x1024x256"  ,G1280x1024x256  , 8},
  {"1280x1024x32K"  ,G1280x1024x32K  ,15},
  {"1280x1024x64K"  ,G1280x1024x64K  ,16},
  {"1280x1024x16M"  ,G1280x1024x16M  ,24},
  {"1280x1024x16M32",G1280x1024x16M32,32},

  {"1600x1200x16"   ,G1600x1200x16   , 4},
  {"1600x1200x256"  ,G1600x1200x256  , 8},
  {"1600x1200x32K"  ,G1600x1200x32K  ,15},
  {"1600x1200x64K"  ,G1600x1200x64K  ,16},
  {"1600x1200x16M"  ,G1600x1200x16M  ,24},
  {"1600x1200x16M32",G1600x1200x16M32,32},

};

int swFindMode(char *mode)
{
  int n;
  for (n=0;n<SUPPORT_MODES;n++)
    if (strcmp(mode,swMI[n].ModeStr)==0) {
      bits_per_pixel = swMI[n].bits;
      #ifdef HAVE_MONO
      if (bits_per_pixel==1) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP4
      if (bits_per_pixel==4) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP8
      if (bits_per_pixel==8) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP15
      if (bits_per_pixel==15) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP16
      if (bits_per_pixel==16) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP24
      if (bits_per_pixel==24) return swMI[n].ModeId;
      #endif
      #ifdef HAVE_BPP32
      if (bits_per_pixel==32) return swMI[n].ModeId;
      #endif
    }
  return TEXT;
}

char * swFindVideoMode(char **mode)
{
  int m;
  while (*mode!=NULL) { 
    m = swFindMode(*mode);
    if (m>0&&vga_hasmode(m)) return *mode;
    mode++;
  }
  return NULL;
}

int swSetupVideo(char *mode)
{
  int m;
  vga_modeinfo * modeinfo;
  m = swFindMode(mode);
  if (m==0) return 0;
  modeinfo = vga_getmodeinfo(m);
  vga_init();
  vga_setmode(m);
  swWidth = modeinfo->width;
  swHeight = modeinfo->height;
  swScanWidth = modeinfo->linewidth;
  switch (bits_per_pixel) {
    #ifdef HAVE_MONO
    case  1: mono_init();  break;
    #endif
    #ifdef HAVE_BPP4
    case  4: bpp4_init();  break;
    #endif
    #ifdef HAVE_BPP8
    case  8: bpp8_init();  break;
    #endif
    #ifdef HAVE_BPP15
    case 15: bpp15_init(); break;
    #endif
    #ifdef HAVE_BPP16
    case 16: bpp16_init(); break;
    #endif
    #ifdef HAVE_BPP24
    case 24: bpp24_init(); break;
    #endif
    #ifdef HAVE_BPP32
    case 32: bpp32_init(); break;
    #endif
  }
  mouse_x = swWidth/2;
  mouse_y = swHeight/2;
  return m;
}

int swClose(void)
{
  vga_setmode(TEXT);
  if (mouse_enable) mouse_close();
  return 1;
}

/*********************************************\
                    MOUSE
\*********************************************/
int mouse_width = 8;
int mouse_height = 12;
int mouse_x = 0;
int mouse_y = 0;
int mice_x  = 0;
int mice_y  = 0;
int mice_update = 0;
int mice_button = 0;
int mouse_memsize = 12 * 8;
int mouse_enable = 1;
int mouse_painted = 0;
unsigned char * mouse_bitmap = NULL;
unsigned char * mouse_screen = NULL;
unsigned char * mouse_paint  = NULL;
MouseHandler mouse_handie=NULL;
void * mouse_pocket = NULL;

int mouse_event(int button,int dx,int dy)
{
  if (!mouse_enable) return 0;
  if (mouse_handie==NULL) return 0;
  return mouse_handie(button,mouse_x+dx,mouse_y+dy,mouse_pocket);
}

int mice_default(int button,int x,int y,void *pocket)
{
  if (x<0) return 0;
  if (y<0) return 0;
  if (x>=(swWidth-mouse_width)) return 0;
  if (y>=(swHeight-mouse_height)) return 0;
  mice_x = x;
  mice_y = y;
  mice_button = button;
  mice_update = 1;
  return 1;
}

int mouse_sethandie(MouseHandler handie,void *pocket)
{
  mouse_handie = handie;
  mouse_pocket = pocket;
}

int mice_init(char *dev, char **xpm)
{
  int r,n;
  if (!mouse_enable) return 0;
  r = mouse_init(dev,vga_getmousetype(),MOUSE_DEFAULTSAMPLERATE);
  if (r!=0) { printf("Disable Mouse!\n"); mouse_enable=0; return 0; }
  mouse_x = swWidth/2;
  mouse_y = swHeight/2;
  mice_grinit(xpm);
  mice_move(mouse_x,mouse_y);
  mouse_seteventhandler(mouse_event);
  mouse_handie = mice_default;
  mouse_setxrange(0,swWidth-1);
  mouse_setyrange(0,swHeight-1);
  return 1;
}

int mice_recover(void)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  mouse_painted = 0;
  return 1;
}

int mice_move(int x,int y)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  getbitmap(x,y,mouse_width,mouse_height,mouse_screen);
  memcpy(mouse_paint,mouse_screen,mouse_memsize);
  miximage(mouse_paint,mouse_bitmap,mouse_width,mouse_height);
  mouse_x = x;
  mouse_y = y;
  bitblt(x,y,mouse_width,mouse_height,mouse_paint);
  mouse_painted = 1;
  return 1;
}

/***************************\
            Font
\***************************/
swFont FONT;

int load_english_font(char *fname)
{
  FILE *fp;
  int n;
  int w = (FONT.eng_w+7)/8;
  int h = FONT.eng_h;
  int s = w*h;

  fp = fopen(fname,"rb");
  if (fp==NULL) return 0;
  for (n=0;n<256;n++) {
    FONT.EFONT[n]=(unsigned char*)malloc(s);
    fread(FONT.EFONT[n],1,s,fp);
  }
  fclose(fp);
  return 1;
}

int swLoadFont(void)
{
  load_english_font(FONT.eng_fname);
  memset(&FONT.info,0,sizeof(FONT.info));
  if (!hbfOpen(FONT.dbcs_path,FONT.dbcs_fname,&FONT.info)) return 0;
  FONT.dbcs_w = FONT.info.mx;
  FONT.dbcs_h = FONT.info.my;
  FONT.CBUFF  = (unsigned char*)malloc(FONT.dbcs_w*FONT.dbcs_h);
  FONT.width  = FONT.eng_w;
  if (FONT.width<FONT.dbcs_w) FONT.width=FONT.dbcs_w;
  FONT.height = FONT.eng_h;
  if (FONT.height<FONT.dbcs_h) FONT.width=FONT.dbcs_h;
  return 1;
}

int swCloseFont(void)
{
  int n;
  for (n=0;n<256;n++) free(FONT.EFONT[n]);
}

unsigned char * swencoding(char *msg)
{
  return hbfGetBitmap(&FONT.info,msg,FONT.CBUFF);
}

int swputs(int x,int y,unsigned char *msg)
{
  unsigned char * dbcs;
  while (*msg) {
    dbcs = swencoding(msg);
    if (dbcs) { /* DBCS */
      drawfont(x,y,FONT.dbcs_w,FONT.dbcs_h,dbcs);
      msg++; msg++; x+=FONT.dbcs_w;
    } else { /* english */
      drawfont(x,y,FONT.eng_w,FONT.eng_h,FONT.EFONT[(int)*msg]);
      msg++; x+=FONT.eng_w;
    }
  }
  return 1;
}

int swStringWidth(unsigned char *msg)
{
  int dbcs;
  int w=0;
  while (*msg) {
    dbcs = swCheckEncoding(msg);
    if (dbcs) { /* DBCS */
      msg++; msg++; w+=FONT.dbcs_w;
    } else { /* english */
      msg++; w+=FONT.eng_w;
    }
  }
  return w;
}

int swputmsg(int x,int y,char *msg)
{
  char buff[1024];
  int n=0,ok=0;

  do {
    if (*msg==10||*msg==0) {
      buff[n]=0;
      swputs(x,y,buff);
      if (*msg==0) ok=1;
      n=0; msg++; y+=(FONT.height+2);
    } else {
      buff[n]=*msg;
      msg++; n++;
    }
  } while (!ok);
  return 1;
}

int swMessageWidth(unsigned char *msg)
{
  char buff[1024];
  int n=0;
  int w=0,x,ok=0;
  
  do {
    if (*msg==10||*msg==0) {
      buff[n]=0;
      x = swStringWidth(buff);
      if (x>w) w=x;
      if (*msg==0) ok=1;
      n=0; msg++;
    } else 
    if (*msg!=0) {
      buff[n]=*msg;
      msg++; n++;
    }
  } while (!ok);
  return w;
}

int swprintf(int x,int y,char *fmt,...)
{
  char buff[2048];
  va_list argptr;
  va_start(argptr,fmt);
  vsprintf(buff,fmt,argptr);
  va_end(argptr);
  swputs(x,y,buff);
  return 1;
}

int hbfOpen(char * path, char * filename,HBFInfo *info)
{
  FILE *      fp;
  char        buff[256];
  int         n,m;
  int         table1;
  int         table2;
  int         count;
  int         count1;
  struct stat attrib;

  strcpy(buff,path);
  strcat(buff,filename);
  fp = fopen(buff,"r");
  if (fp==NULL) return 0;

  /* Get Font Information */
  while (fgets(buff,119,fp)!=NULL)
    if (strncmp(buff,"HBF_BITMAP_BOUNDING_BOX",23)==0) break;
  sscanf(buff,"%*s%d%d",&info->mx,&info->my);
  info->bytes = info->mx * info->my / 8;

  /* get byte 2 ranges */
  while (fgets(buff,119,fp)!=NULL)
    if (strncmp(buff,"HBF_START_BYTE_2_RANGES",23)==0) break;
  sscanf(buff,"%*s%d",&info->byte2range);

  /* Load Byte 2 Ranges */
  info->B2Info = (Byte2Info*)malloc(sizeof(Byte2Info)*(info->byte2range));
  info->totalrange = 0;
  for (n=0;n<info->byte2range;n++) {
    while (fgets(buff,119,fp)!=NULL)
      if (strncmp(buff,"HBF_BYTE_2_RANGE",16)==0) break;
    sscanf(buff,"%*s%i-%i",&info->B2Info[n].l,&info->B2Info[n].h);
    info->B2Info[n].a = info->B2Info[n].h - info->B2Info[n].l + 1;
    info->totalrange += info->B2Info[n].a;
  }

  /* Get Code Ranges Information */
  while (fgets(buff,119,fp)!=NULL)
    if (strncmp(buff,"HBF_START_CODE_RANGES",21)==0) break;
  sscanf(buff,"%*s%d",&info->segs);
  info->SEG = (HBFRange *)malloc(sizeof(HBFRange)*(info->segs));

  /* Read Segments Information */
  for (n=0;n<info->segs;n++) {
    while (fgets(buff,119,fp)!=NULL)
      if (strncmp(buff,"HBF_CODE_RANGE",14)==0) break;
    sscanf(buff,"%*s%x-%x%s%d",
      &info->SEG[n].l,
      &info->SEG[n].h,
      info->SEG[n].bmfname,
      &info->SEG[n].offset
    );

    count1 = 0;
    count  = info->SEG[n].l & 0xff;
    for (table1=0;table1<256;table1++,count=(count+1)%256) {
      info->SEG[n].address[count] = -1;
      for (table2=0;table2<info->byte2range;table2++) {
        if ((count>=info->B2Info[table2].l)&&(count<=info->B2Info[table2].h))
          info->SEG[n].address[count] = count1;
      }
      if (info->SEG[n].address[count]!=-1) count1++;
    }
    if (n>0) {
      for (m=0;m<n;m++)
        if (strcmp(info->SEG[m].bmfname,info->SEG[n].bmfname)==0)
          info->SEG[n].fp = info->SEG[m].fp;
    }
    if (info->SEG[n].fp==NULL) {
      strcpy(buff,path);
      strcat(buff,info->SEG[n].bmfname);
      info->SEG[n].fp = fopen(buff,"rb");
      if (info->SEG[n].fp==NULL) return 0;
    }
  }
  fclose(fp);
  return 1;
}

int swCheckEncoding(char *str)
{
  unsigned char * p = (unsigned char *)str;
  unsigned int lsb = (unsigned int)p[1];
  unsigned int msb = (unsigned int)p[0];
  unsigned int ch = msb<<8 | lsb;
  long address;
  int offset;
  int n;

  /* looking for the segment of this char */
  for (n=0;n<FONT.info.segs;n++)
    if (ch>=FONT.info.SEG[n].l&&ch<=FONT.info.SEG[n].h) {
      /* Gotcha */
      offset = FONT.info.SEG[n].address[lsb];
      if (offset==-1) return 0;
      return 1;
    }

  return 0;
}

unsigned char * hbfGetBitmap(HBFInfo *info,char *str,unsigned char *addr)
{
  unsigned char * p = (unsigned char *)str;
  unsigned int lsb = (unsigned int)p[1];
  unsigned int msb = (unsigned int)p[0];
  unsigned int ch = msb<<8 | lsb;
  long address;
  int offset;
  int n;

  /* looking for the segment of this char */
  for (n=0;n<info->segs;n++)
    if (ch>=info->SEG[n].l&&ch<=info->SEG[n].h) {
      /* Gotcha */
      offset = info->SEG[n].address[lsb];
      if (offset==-1) return NULL;
      ch -= info->SEG[n].l;
      address = info->SEG[n].offset + (info->totalrange*(ch>>8)+offset)*info->bytes;
      fseek(info->SEG[n].fp,address,SEEK_SET);
      fread(addr,info->bytes,1,info->SEG[n].fp);
      return addr;
    }

  return NULL;
}

int hbfClose(HBFInfo *info)
{
  int n;
//  for (n=0;n<info->segs;n++) fclose(info->SEG[n].fp);
  free(info->B2Info);
  free(info->SEG);
}

/************************************\
        Star Window Manager
\************************************/

void swScaleColor(swCOLOR * color,float scale)
{
  #define MU(v) v = (int)(((float)v)*scale); if (v>255) v=255
  MU(color->r);
  MU(color->g);
  MU(color->b);
  #undef  MU
}

int swIsInside(int x,int y,swRect *RECT)
{
  if (x<RECT->x1) return 0;
  if (y<RECT->y1) return 0;
  if (x>RECT->x2) return 0;
  if (y>RECT->y2) return 0;
  return 1;
}

int swDrawBorder(swRect * RECT,swCOLOR * Lite,swCOLOR * Vampire)
{
  /* Draw Lite Side */
  setcolor(Lite->r,Lite->g,Lite->b);
  drawline(RECT->x1,RECT->y1,RECT->x2,RECT->y1);
  drawline(RECT->x1,RECT->y1+1,RECT->x1,RECT->y2-1);
  /* Draw Vampire */
  setcolor(Vampire->r,Vampire->g,Vampire->b);
  drawline(RECT->x1,RECT->y2,RECT->x2,RECT->y2);
  drawline(RECT->x2,RECT->y1+1,RECT->x2,RECT->y2-1);
  return 1;
}

int swDrawBox(swRect * RECT,swCOLOR * color,int up_down)
{
  swCOLOR Lite =  *color;
  swCOLOR Vampire = *color;
  swRect FX = *RECT;
  
  (up_down) ? swScaleColor(&Lite,1.4):swScaleColor(&Lite,0.6);
  (up_down) ? swScaleColor(&Vampire,0.6):swScaleColor(&Vampire,1.4);
  
  swDrawBorder(&FX,&Lite,&Vampire);
  FX.x1++; FX.y1++; FX.x2--; FX.y2--;
  swDrawBorder(&FX,&Lite,&Vampire);
  FX.x1++; FX.y1++; FX.x2--; FX.y2--;
  setcolor(color->r,color->g,color->b);
  fillbox(FX.x1,FX.y1,FX.x2-FX.x1+1,FX.y2-FX.y1+1);

  return 1;
}

/**************************************\
                Button
\**************************************/
int swDrawButton(
      swRect *RECT,
      swCOLOR *color,
      swCOLOR *text,
      int up_down,
      char *fmt,...
    )
{
  int w;
  int x,y;
  char buff[256];
  va_list argptr;

  va_start(argptr,fmt);
  vsprintf(buff,fmt,argptr);
  va_end(argptr);

  /* Calculate width */
  x = RECT->x2-RECT->x1-3;
  y = RECT->y2-RECT->y1-3;
  w = swStringWidth(buff);
  x -= w;           x/=2; x+=RECT->x1+2;
  y -= FONT.height; y/=2; y+=RECT->y1+2;

  /* Draw  */
  swDrawBox(RECT,color,up_down);
  setforeground(text->r,text->g,text->b);
  setbackground(color->r,color->g,color->b);
  swputs(x,y,buff);

  return 1;
}

/**************************************\
            Key Functions
\**************************************/
int swConsole = 0;
static struct termios swOldTerm;
static fd_set swKEY;

int swSetupKey(char * dev_console)
{
  struct termios buf;
  
  swConsole = open(dev_console,O_RDWR);
  tcgetattr(swConsole,&swOldTerm);

  buf = swOldTerm;
  
//  cfmakeraw(&buf);

  buf.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
  buf.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  buf.c_cflag &= ~(CSIZE | PARENB);
  buf.c_cflag |= CS8;
  buf.c_oflag &= ~(OPOST);

  buf.c_cc[VMIN] = 1;
  buf.c_cc[VTIME] = 0;
  if (tcsetattr(swConsole,TCSAFLUSH,&buf)<0) return 0;

  FD_ZERO(&swKEY);
  FD_SET(swConsole,&swKEY);
  
  
}

int swCloseKey(void)
{
//  write(swConsole,"\033c",2);
  close(swConsole);
}

int swIsKeyAvailable(void)
{
  struct timeval tv;
  unsigned char c;
  fd_set rset;
  
  rset = swKEY;
  tv.tv_sec = tv.tv_usec = 0;
  if (select(FD_SETSIZE,&rset,NULL,NULL,&tv)==-1) return 0;
  if (!FD_ISSET(swConsole,&rset)) return 0;
  return 1;
}

int swGetKey(void)
{
  unsigned char c;
  if (!swIsKeyAvailable()) return -1;
  read(swConsole,&c,sizeof(char));
  return (int)c;
}

int swGetCH(void)
{
  unsigned char c;
  read(swConsole,&c,sizeof(char));
  return (int)c;
}

void swBeep(void)
{
//  ioctl(swConsole,KDMKTONE,0x100637);
}

/**************************************\
            Draw Message Box
\**************************************/
unsigned char * swDrawMessageBox(swRect * RECT,char *title,char *msg)
{
  int H=0;
  int W,w;
  char *p = msg;
  swCOLOR f,b;
  unsigned char * bkgnd;
  
  while (*p) {
    if (*p=='\n') H++;
    p++;
  }
  H++;
  H *= (FONT.height+2);
  H += FONT.height+4; /* title */
  H += FONT.height*3/2;
  H += FONT.height+8;
  H += 4; /* border */
  
  w = swStringWidth(title);
  W = swMessageWidth(msg);
  if (w>W) W=w;
  W+=8;
  
  RECT->x1 = (swWidth  - W)/2;
  RECT->y1 = (swHeight - H)/2;
  RECT->x2 = RECT->x1 + W - 1;
  RECT->y2 = RECT->y1 + H - 1;
  
  bkgnd = create_buffer(RECT->x2-RECT->x1+1,RECT->y2-RECT->y1+1);
  getbitmap(
    RECT->x1,RECT->y1,
    RECT->x2-RECT->x1+1,RECT->y2-RECT->y1+1,
    bkgnd
  );  
  set_swcolor(f,192,192,192);
  swDrawBox(RECT,&f,1);
  
  setcolor(255,0,255);
  fillbox(RECT->x1+2,RECT->y1+2,W-4,FONT.height+4);
  setforeground(0,0,255);
  setbackground(255,0,255);
  swputs(RECT->x1+(W-w)/2,RECT->y1+4,title);

  setforeground(0,0,255);
  setbackground(192,192,192);
  swputmsg(RECT->x1+4,RECT->y1+4+FONT.height*3/2,msg);

  return bkgnd;
}

/**************************************\
             Message Box
                   
                   W+8
 +------------------------------------+
 |                TITLE               | H+4
 +------------------------------------+
 > h/2
 |                                    |
 |               MESSAGES             | H+2
 |                                    |
 |                                    |
 > h/2
 |               | btn |              | H+8
 > h/2
 +------------------------------------+ 
 
\**************************************/
int swMessageBox(char *title,char *btn,char *fmt,...)
{
  char msg[4096];
  swRect RT;
  swRect BTN;
  swCOLOR b;
  swCOLOR t;
  int w,drawstate=0;
  int ok,c;
  unsigned char * bkgnd;
  va_list argptr;
  
  va_start(argptr,fmt);
  vsprintf(msg,fmt,argptr);
  va_end(argptr);
  
  mice_recover();
  w = swStringWidth(btn);
  bkgnd = swDrawMessageBox(&RT,title,msg);
  BTN.x1 = RT.x1;  BTN.x2 = RT.x2-RT.x1-3;
  BTN.x1+= (BTN.x2-w-16)/2;
  BTN.x2 = BTN.x1 + w + 16;
  
  BTN.y1 = RT.y2 - 10 - FONT.height*3/2;
  BTN.y2 = BTN.y1 + FONT.height + 8;

  set_swcolor(b,192,192,192);
  set_swcolor(t,255,0,0);
  swDrawButton(&BTN,&b,&t,1,btn);
  
  mice_move(mouse_x,mouse_y);
  
  ok=0;
  
  do {
    if (mouse_update()) {
      if (swIsInside(mice_x,mice_y,&BTN)) {
        if (mice_button&MOUSE_LEFTBUTTON) {
          if (drawstate!=2) {
            mice_recover();
            set_swcolor(b,224,0,224);
            set_swcolor(t,0,0,255);
            swDrawButton(&BTN,&b,&t,0,btn); 
            drawstate=2;
            mice_move(mice_x,mice_y);
            ok = 1;
          }
        } else 
        if (drawstate!=1) {
          mice_recover();
          set_swcolor(b,224,0,224);
          set_swcolor(t,0,0,255);
          swDrawButton(&BTN,&b,&t,1,btn); 
          drawstate=1;
          mice_move(mice_x,mice_y);
        }
      } else {
        if (drawstate!=0) {
          mice_recover();
          set_swcolor(b,192,192,192);
          set_swcolor(t,255,0,0);
          swDrawButton(&BTN,&b,&t,1,btn); 
          drawstate=0;
          mice_move(mice_x,mice_y);
        }
      }
      if (mice_update) mice_move(mice_x,mice_y);
    }
    c=swGetKey();
    if (c==13||c==' '||c==10||c==9) ok=1;
  } while (!ok);
  
  mice_recover();  
  bitblt(RT.x1,RT.y1,RT.x2-RT.x1+1,RT.y2-RT.y1+1,bkgnd);
  mice_move(mice_x,mice_y);

  free(bkgnd);
  return 1;
}

/***************************************\
          Message Box - Yes/No
\***************************************/
int swNoticeYesNo(char *title,char *yes,char *no,char *fmt,...)
{
  char msg[4096];
  swRect RT;
  swRect YES;
  swRect NO;
  swCOLOR b;
  swCOLOR t;
  int w,drawstate=0;
  int ok,c;
  unsigned char * bkgnd;
  va_list argptr;
  
  va_start(argptr,fmt);
  vsprintf(msg,fmt,argptr);
  va_end(argptr);
  
  mice_recover();
  w = swStringWidth(yes);
  bkgnd = swDrawMessageBox(&RT,title,msg);
  YES.x1 = (RT.x2-RT.x1)/4 - (w + 8)/2;  
  YES.x1+= RT.x1;
  YES.x2 = YES.x1 + w + 8;
  
  YES.y1 = RT.y2 - 10 - FONT.height*3/2;
  YES.y2 = YES.y1 + FONT.height + 8;

  set_swcolor(b,192,192,192);
  set_swcolor(t,255,0,0);
  swDrawButton(&YES,&b,&t,1,yes);

  NO = YES;
  w = swStringWidth(no);
  NO.x1 = (RT.x2-RT.x1)*3/4 - (w + 8)/2; 
  NO.x1+= RT.x1;
  NO.x2 = NO.x1 + w + 8;
  swDrawButton(&NO,&b,&t,1,no);
  
  mice_move(mouse_x,mouse_y);
  
  ok=0;
  
  do {
    if (mouse_update()) {
      if (swIsInside(mice_x,mice_y,&YES)) {
        if (mice_button&MOUSE_LEFTBUTTON) {
          if (drawstate!=2) {
            mice_recover();
            set_swcolor(b,224,0,224);
            set_swcolor(t,0,0,255);
            swDrawButton(&YES,&b,&t,0,yes); 
            drawstate=2;
            mice_move(mice_x,mice_y);
            ok = 2;
          }
        } else 
        if (drawstate!=1) {
          mice_recover();
          set_swcolor(b,224,0,224);
          set_swcolor(t,0,0,255);
          swDrawButton(&YES,&b,&t,1,yes); 
          drawstate=1;
          mice_move(mice_x,mice_y);
        }
      } else 
      if (swIsInside(mice_x,mice_y,&NO)) {
        if (mice_button&MOUSE_LEFTBUTTON) {
          if (drawstate!=2) {
            mice_recover();
            set_swcolor(b,224,0,224);
            set_swcolor(t,0,0,255);
            swDrawButton(&NO,&b,&t,0,no); 
            drawstate=2;
            mice_move(mice_x,mice_y);
            ok = 1;
          }
        } else 
        if (drawstate!=1) {
          mice_recover();
          set_swcolor(b,224,0,224);
          set_swcolor(t,0,0,255);
          swDrawButton(&NO,&b,&t,1,no); 
          drawstate=1;
          mice_move(mice_x,mice_y);
        }
      } else {
        if (drawstate!=0) {
          mice_recover();
          set_swcolor(b,192,192,192);
          set_swcolor(t,255,0,0);
          swDrawButton(&YES,&b,&t,1,yes); 
          swDrawButton(&NO,&b,&t,1,no); 
          mice_move(mice_x,mice_y);
          drawstate = 0;
        }
      }
      if (mice_update) mice_move(mice_x,mice_y);
    }
    c=swGetKey();
    if (c==13||c==' '||c==10||c==9||c=='n'||c=='N') ok=1;
    if (c=='Y'||c=='y') ok=2;
  } while (!ok);
  
  mice_recover();  
  bitblt(RT.x1,RT.y1,RT.x2-RT.x1+1,RT.y2-RT.y1+1,bkgnd);
  mice_move(mice_x,mice_y);

  free(bkgnd);

  return ok-1; 
}

/************************************\
                Menu
\************************************/
int swDoMenu(int x,int y,int gravity,swMenu * MENU)
{
  int w=0,n,ww;
  int h;
  swRect RT;
  swRect BTN;
  unsigned char * bkgnd;
  int ID;
  swCOLOR bg;
  swCOLOR fg;
  int ok=0;

  /* Calculate Dimension */
  h = MENU->size * (FONT.height+6);
  for (n=0;n<MENU->size;n++) {
    ww = swStringWidth(MENU->menu[n].menu);
    if (ww>w) w=ww;
  }
  w+=6;

  switch (gravity) {
    case GRAVITY_NW: RT.x1 = x;     RT.y1 = y;         break;
    case GRAVITY_NE: RT.x1 = x-w+1; RT.y1 = y;         break;
    case GRAVITY_SW: RT.x1 = x;     RT.y1 = y - h + 1; break;
    case GRAVITY_SE: RT.x1 = x-w+1; RT.y1 = y - h + 1; break;
  }

  RT.x2 = RT.x1 + w - 1;
  RT.y2 = RT.y1 + h - 1;

  mice_recover();
  /* save background */
  bkgnd = create_buffer(w,h);
  getbitmap(RT.x1,RT.y1,w,h,bkgnd);  
  
  /* Draw Menu */
  set_swcolor(bg,192,192,192);
  set_swcolor(fg,0,0,255);
  for (n=0;n<MENU->size;n++) {
    BTN.x1 = RT.x1;
    BTN.x2 = RT.x2; 
    BTN.y1 = RT.y1 + n * (FONT.height+6);
    BTN.y2 = BTN.y1 + FONT.height + 5;
    swDrawButton(&BTN,&bg,&fg,1,MENU->menu[n].menu);
  }
  mice_move(mice_x,mice_y);
  ID = -1;
  do {
    if (mouse_update()) {
      if (mice_update) mice_move(mice_x,mice_y);
      if (swIsInside(mice_x,mice_y,&RT)) {
        n = (mice_y - RT.y1)/(FONT.height+6);
        if (ID<0||n!=ID) mice_recover();
        if (ID>=0&&n!=ID) {
          set_swcolor(bg,192,192,192);
          set_swcolor(fg,0,0,255);
          BTN.x1 = RT.x1;
          BTN.x2 = RT.x2; 
          BTN.y1 = RT.y1 + ID * (FONT.height+6);
          BTN.y2 = BTN.y1 + FONT.height + 5;
          swDrawButton(&BTN,&bg,&fg,1,MENU->menu[ID].menu);
        }
        if (n!=ID) {
          set_swcolor(bg,192,0,192);
          set_swcolor(fg,255,255,0);
          BTN.x1 = RT.x1;
          BTN.x2 = RT.x2; 
          BTN.y1 = RT.y1 + n * (FONT.height+6);
          BTN.y2 = BTN.y1 + FONT.height + 5;
          swDrawButton(&BTN,&bg,&fg,1,MENU->menu[n].menu);
        }
        if (ID<0||n!=ID) mice_move(mice_x,mice_y);
        ID = n;
        if (mice_button&(MOUSE_LEFTBUTTON|MOUSE_RIGHTBUTTON)) ok=1;
      } else {
        if (ID>=0) {
          mice_recover();
          set_swcolor(bg,192,192,192);
          set_swcolor(fg,0,0,255);
          BTN.x1 = RT.x1;
          BTN.x2 = RT.x2; 
          BTN.y1 = RT.y1 + ID * (FONT.height+6);
          BTN.y2 = BTN.y1 + FONT.height + 5;
          swDrawButton(&BTN,&bg,&fg,1,MENU->menu[n].menu);
          mice_move(mice_x,mice_y);
          ID = -1;
        }
        if (mice_button&(MOUSE_LEFTBUTTON|MOUSE_RIGHTBUTTON)) ok=1;
      }
    }
    if (swIsKeyAvailable()) {
    }
  } while (!ok);

  mice_recover();  
  bitblt(RT.x1,RT.y1,w,h,bkgnd);
  mice_move(mice_x,mice_y);

  free(bkgnd);

  return (ID<0) ? -1 : MENU->menu[ID].id;
}

/****************************************\
             TTY Functions
\****************************************/
int open_pty(char *pty)
{
  int tty;
  char *ptr1;
  char *ptr2;
  strcpy(pty,"/dev/ptyXX");
  
  for (ptr1="pqrst";*ptr1;ptr1++) {
    pty[8]=*ptr1;
    for (ptr2="0123456789abcdef";*ptr2;ptr2++) {
      pty[9]=*ptr2;
      tty = open(pty,O_RDWR);
      if (tty!=-1) return tty;
    }
  }
  return -1;
}

pid_t open_tty(char *ptyname,int col,int row,int login)
{
  char ttyname[16];
  pid_t pid;
  struct passwd *userpd;
  int n;
  int tty_fd;
  struct winsize ws;
  
  strcpy(ttyname,ptyname);
  ttyname[5] = 't';
  
  pid = fork();
  if (pid==-1) return pid;
  if (pid!=0)  return pid;
  
  /* close all opened files */
  for (n=3;n<256;n++) close(n);

  setsid();
  tty_fd = open(ttyname,O_RDWR);
  if (tty_fd==-1) { close(tty_fd); exit(-1); }

  ws.ws_row = row;
  ws.ws_col = col;
  
  ioctl(tty_fd,TIOCSCTTY,0);
  ioctl(tty_fd,TIOCSWINSZ,&ws);

  seteuid(getuid());
  dup2(tty_fd,fileno(stdin));
  dup2(tty_fd,fileno(stdout));
  dup2(tty_fd,fileno(stderr));
  close(tty_fd);

  userpd = getpwuid(getuid());
  if (userpd==NULL||login) {
    /* use login - /bin/login */
    execlp("/bin/login","/bin/login",(char*)0);
  } else
    execlp(userpd->pw_shell,userpd->pw_shell,(char*)0);

  exit(-1);
  return pid;
}


